/*
 * Copyright (c) 2007 Kathryn Huxtable
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: PropertiesBundle.java 578 2007-04-07 15:30:39Z huxtable $
 */
package org.kathrynhuxtable.middleware.shibshim.util.properties;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import org.apache.log4j.Logger;

/**
 * A singleton class to store the property bundle where it can be retrieved
 * globally.
 */
public class PropertiesBundle {
    /**
     * Logger for error/information/debug messages.
     */
    private static Logger     log        = Logger.getLogger(PropertiesBundle.class.getName());

    /**
     * Properties read from the properties file.
     */
    private static Properties properties = null;

    /**
     * Initialize the properties, reading them from the file. If the properties
     * have already been initialized, this does nothing.
     * 
     * @param propertiesLocation
     *            the file containing the properties.
     * @throws PropertiesBundleException
     *             if the file does not exist or if the properties cannot be
     *             loaded from the file.
     */
    public static void init(String propertiesLocation) throws PropertiesBundleException {
        if (properties == null) {
            log.info("Loading properties bundle from " + propertiesLocation);
            // Get the properties location as a URL resource.
            URL resource = null;
            try {
                resource = new URL(propertiesLocation);
            } catch (MalformedURLException e) {
                resource = PropertiesBundle.class.getResource(propertiesLocation);
            }
            if (resource == null) {
                log.fatal("Properties file could not be found at the specified location: " + propertiesLocation);
                throw new PropertiesBundleException("Properties file could not be found at the specified location: " + propertiesLocation);
            }

            // Load the properties file
            Properties props = new Properties();
            try {
                props.load(resource.openStream());
            } catch (FileNotFoundException e) {
                throw new PropertiesBundleException(e);
            } catch (IOException e) {
                throw new PropertiesBundleException(e);
            }

            properties = props;
        }
    }

    /**
     * Returns a set view of the property names.
     * 
     * @return the set of property names.
     */
    public static Set getPropertyNames() {
        return properties.keySet();
    }

    /**
     * Get a string value for a property.
     * 
     * @param label
     *            a descriptive label for what the property specifies.
     * @param propertyName
     *            the name of the property.
     * @return the value of the property.
     * @throws PropertiesBundleException
     *             if the property is not present.
     */
    public static String getString(String label, String propertyName) throws PropertiesBundleException {
        String prop = properties.getProperty(propertyName);
        if (prop == null || prop.equals("")) {
            log.fatal("No " + label + " specified. Use property (" + propertyName + ")");
            throw new PropertiesBundleException("Invalid configuration data supplied.");
        }
        return prop;
    }

    /**
     * Get a string value for a property.
     * 
     * @param label
     *            a descriptive label for what the property specifies.
     * @param propertyName
     *            the name of the property.
     * @param defaultValue
     *            the default value if the property is not required and not
     *            present.
     * @return the value of the property.
     * @throws PropertiesBundleException
     *             if the property is required and not present.
     */
    public static String getString(String label, String propertyName, String defaultValue) throws PropertiesBundleException {
        String prop = properties.getProperty(propertyName);
        if (prop == null || prop.equals("")) {
            prop = defaultValue;
        }
        return prop;
    }

    /**
     * Get an integer value for a property.
     * 
     * @param label
     *            a descriptive label for what the property specifies.
     * @param propertyName
     *            the name of the property.
     * @return the value of the property.
     * @throws PropertiesBundleException
     *             if the property is not present.
     */
    public static int getInteger(String label, String propertyName) throws PropertiesBundleException {
        return Integer.parseInt(getString(label, propertyName));
    }

    /**
     * Get an integer value for a property.
     * 
     * @param label
     *            a descriptive label for what the property specifies.
     * @param propertyName
     *            the name of the property.
     * @param defaultValue
     *            the default value if the property is not required and not
     *            present.
     * @return the value of the property.
     * @throws PropertiesBundleException
     *             if the property is required and not present.
     */
    public static int getInteger(String label, String propertyName, int defaultValue) throws PropertiesBundleException {
        return Integer.parseInt(getString(label, propertyName, Integer.toString(defaultValue)));
    }

    /**
     * @return a map representing all the properties beginning with an initial
     *         string mapped to their values.
     * @throws PropertiesBundleException
     *             if there are no properties matching the initial string.
     */
    public static Map getMap(String label, String propertyName) throws PropertiesBundleException {
        // Parse the various "propertyName.*" properties to get the
        // attributeMap.
        Map map = new HashMap();
        Iterator iter = PropertiesBundle.getPropertyNames().iterator();
        while (iter.hasNext()) {
            String name = (String) iter.next();
            if (name.startsWith(propertyName)) {
                String attribute = name.substring(propertyName.length());
                String value = properties.getProperty(name);
                map.put(attribute.toLowerCase(), value);
            }
        }
        if (map.size() == 0) {
            log.fatal("No " + label + " specified. Use properties (" + propertyName + "*)");
            throw new PropertiesBundleException("Invalid configuration data supplied.");
        }
        return map;
    }

    /**
     * @return a map representing all the properties beginning with an initial
     *         string mapped to their values.
     * @throws PropertiesBundleException
     *             if the properties have not been initialized.
     */
    public static Map getMap(String label, String propertyName, Map defaultValue) throws PropertiesBundleException {
        // Parse the various "propertyName.*" properties to get the
        // attributeMap.
        Map map = new HashMap();
        Iterator iter = PropertiesBundle.getPropertyNames().iterator();
        while (iter.hasNext()) {
            String name = (String) iter.next();
            if (name.startsWith(propertyName)) {
                String attribute = name.substring(propertyName.length());
                String value = properties.getProperty(name);
                map.put(attribute.toLowerCase(), value);
            }
        }
        if (map.size() == 0) {
            return defaultValue;
        }
        return map;
    }
}
