/*
 Copyright (c) 2007 Kathryn Huxtable

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

 $Id: ACSServlet.java 1156 2007-04-15 01:32:07Z kathryn5 $
 */
package org.kathrynhuxtable.middleware.shibshim.web.acs;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.log4j.Logger;
import org.apache.log4j.MDC;
import org.doomdark.uuid.UUIDGenerator;
import org.kathrynhuxtable.middleware.shibshim.attributes.AttributeFormatter;
import org.kathrynhuxtable.middleware.shibshim.util.AESCrypt;
import org.kathrynhuxtable.middleware.shibshim.util.AESCryptException;
import org.kathrynhuxtable.middleware.shibshim.util.Base64;
import org.kathrynhuxtable.middleware.shibshim.util.RSASignature;
import org.kathrynhuxtable.middleware.shibshim.util.RSASignatureException;
import org.kathrynhuxtable.middleware.shibshim.util.TimeUtil;
import org.kathrynhuxtable.middleware.shibshim.util.properties.PropertiesBundle;
import org.kathrynhuxtable.middleware.shibshim.util.properties.PropertiesBundleException;

/**
 * <p>
 * The <b>ACSServlet</b> is the servlet used to retrieve attributes from
 * Shibboleth and pass them on to the Shibboleth Shim filter.
 */
public class ACSServlet extends HttpServlet {
    private static final Logger log        = Logger.getLogger(ACSServlet.class.getName());

    private String              privateKey = null;
    private String              cryptKey   = null;

    /**
     * {@inheritDoc}
     */
    public void init(ServletConfig config) throws ServletException {
        MDC.put("serviceId", "[Attr]  Core");

        // Get the properties file name.
        String propertiesLocation = config.getServletContext().getInitParameter("propertiesFile");
        if (propertiesLocation == null) {
            propertiesLocation = "/shibshim.properties";
        }

        // Load the properties bundle from the properties file.
        // Get the values.
        try {
            PropertiesBundle.init(propertiesLocation);

            privateKey = readFile(PropertiesBundle.getString("private key", "privateKey"));
            cryptKey = readFile(PropertiesBundle.getString("encryption key", "cryptKey"));
        } catch (PropertiesBundleException e) {
            throw new ServletException(e);
        }

        log.info("ACSServlet initialized.");
    }

    /**
     * {@inheritDoc}
     */
    public void doGet(HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException {
        handleRequest(request, response);
    }

    /**
     * {@inheritDoc}
     */
    public void doPost(HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException {
        handleRequest(request, response);
    }

    /**
     * Common code to handle GET and POST requests. Realistically, we should
     * only get POST requests. This gets the attributes passed by Shibboleth,
     * formats them in a simple XML document, signs them with a private key, and
     * POST redirects to the Assertion Consumer Service in the Shibboleth Shim
     * filter.
     * 
     * @param request
     * @param response
     * @throws IOException
     * @throws ServletException
     */
    public void handleRequest(HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException {
        MDC.put("serviceId", "[Attr]  " + UUIDGenerator.getInstance().generateRandomBasedUUID());
        MDC.put("remoteAddr", request.getRemoteAddr());
        log.info("Handling request");

        try {
            String lastAuthTime = TimeUtil.formatToGMT(new Date());

            // Get the Shibboleth IdP. There may be a better way to get this.
            String url = request.getHeader("Referer");
            if (url != null) {
                url = url.replaceFirst("\\?.*$", "");
            }

            // For now, use the session ID of the Shibboleth Shim server as the
            // handle for the user session.
            String handle = request.getSession().getId();

            // Retrieve the Shibboleth attributes and assemble them into an XML
            // document (the assertion).
            Map attributes = parseAttributes(request);
            String assertion = AttributeFormatter.formatAssertion(attributes, lastAuthTime, url, handle);

            // Generate the signature for the assertion.
            String signature = null;
            try {
                signature = Base64.encodeBytes(RSASignature.sign(assertion.getBytes(), privateKey));
            } catch (RSASignatureException e) {
                request.setAttribute("errorText", "Unable to sign attribute assertion: " + e);
                request.getRequestDispatcher("/WEB-INF/jsp/error.jsp").forward(request, response);
            }

            // Encrypt and base64 encode the assertion.
            String base64CipherText = null;
            try {
                byte[] cipherText = AESCrypt.encrypt(assertion.getBytes(), cryptKey);
                base64CipherText = Base64.encodeBytes(cipherText);
            } catch (AESCryptException e) {
                request.setAttribute("errorText", "Unable to encrypt attribute assertion: " + e);
                request.getRequestDispatcher("/WEB-INF/jsp/error.jsp").forward(request, response);
            }

            // Parse the ACS for the Shibboleth Shim filter out of the extra
            // path info in the request.
            String filterACS = request.getPathInfo();
            if (filterACS == null) {
                filterACS = "";
            } else {
                if (filterACS.charAt(0) == '/') {
                    filterACS = filterACS.substring(1);
                }
                if (filterACS.startsWith("https/")) {
                    filterACS = filterACS.replaceFirst("https/", "https://");
                } else if (filterACS.startsWith("http/")) {
                    filterACS = filterACS.replaceFirst("http/", "http://");
                }
            }

            // Set variables and POST our redirect.
            request.setAttribute("filterACS", filterACS);
            request.setAttribute("target", request.getParameter("target"));
            request.setAttribute("assertion", base64CipherText);
            request.setAttribute("signature", signature);

            request.getRequestDispatcher("/WEB-INF/jsp/post.jsp").forward(request, response);
        } finally {
            log.info("Finished request");
        }
    }

    /**
     * Retrieves the user attributes in a Map&lt;String, Set&lt;String&gt;&gt;
     * (in Java5 notation).
     * 
     * @param request
     *            The request containing the attributes.
     * 
     * @return user attributes in a map of attribute names to value sets.
     */
    private Map parseAttributes(HttpServletRequest request) {
        log.debug("Parsing user attributes");

        Map attributes = new HashMap();

        // Special check for edpPersonPrincipalName, which is mapped to remote
        // user.
        if (request.getRemoteUser() != null && request.getRemoteUser().length() > 0) {
            Set values = new HashSet();
            values.add(request.getRemoteUser());
            attributes.put("Shib-EP-EduPersonPrincipalName", values);
        }

        // Get the rest of the Shibboleth attributes.
        Enumeration attributeEnum = request.getHeaderNames();
        while (attributeEnum.hasMoreElements()) {
            String attributeName = (String) attributeEnum.nextElement();
            // TODO Need to have more flexible header selection.
            if (!attributeName.startsWith("Shib-")) {
                continue;
            }

            Set values = new HashSet();
            Enumeration v = request.getHeaders(attributeName);
            while (v.hasMoreElements()) {
                String value = (String) v.nextElement();

                value = value.replaceAll("([^\\\\]);", "$1\0");
                value = value.replaceAll("\\;", ";");

                String[] elements = value.split("\0");
                for (int i = 0; i < elements.length; i++) {
                    values.add(elements[i]);
                }
            }

            attributes.put(attributeName, values);
        }
        return attributes;
    }

    private String readFile(String filename) throws ServletException {
        // Read the private key file.
        try {
            BufferedReader reader = new BufferedReader(new FileReader(new File(filename)));
            StringBuffer privateKeyBuf = new StringBuffer();
            String buffer = null;
            while ((buffer = reader.readLine()) != null) {
                privateKeyBuf.append(buffer).append('\n');
            }
            reader.close();
            String text = privateKeyBuf.toString();
            return text;
        } catch (FileNotFoundException e) {
            log.fatal("File not found: " + privateKey);
            throw new ServletException(e);
        } catch (IOException e) {
            log.fatal("IO Exception: " + privateKey);
            throw new ServletException(e);
        }
    }
}
